#ifdef SCCS
static char sccsid[]="@(#)dial.c	1.2 Stellar 89/03/01";
#endif
/*
			Copyright 1989 by
			Stellar Computer Inc.
			All Rights Reserved
	
	This software comprises unpublished confidential information of
	Stellar Computer Inc. and may not be used, copied or made
	available to anyone, except in accordance with the license
	under which it is furnished.
*/
/*
 * LUI Dial Box
 */

#include <stdio.h>
#include <math.h>

#define LUI_DIAL_DECLARE
#include "lui.h"

static LUI_FNCP LUI_DialFunction;

/******************************************************************************/
LUI_DialBoxOpen(name, parent, x, y, border, func)
    char *name;
    Window parent;
    int x, y, border;
    LUI_FNCP func;
{
    LUI_DIAL_BOX *dial_box;
    Window LUI_DialBoxCreateWindow();
    int LUI_DialBoxDraw();

    /* Check for errors */
    if (LUI_DialBox) {
	fprintf(stderr, "LUI_DialBoxOpen: DialBox %s still open!\n", LUI_DialBox->name);
	return(0);
    }

    /* Initialize placement variables */
    LUI_DialBoxWidth  = 0;
    LUI_DialBoxHeight = 0;

    /* Create the dial box window */
    LUI_DialBoxWindow = LUI_CreateWindow(parent, 0, 0);

    /* Allocate the dial box */
    if (!(dial_box = (LUI_DIAL_BOX *)malloc(sizeof(LUI_DIAL_BOX)))) {
	fprintf(stderr, "LUI_DialBoxOpen: can't allocate dial box\n");
	return(0);
    }

    /* Set Current dial box and function */
    LUI_DialBox      = dial_box;
    LUI_DialFunction = func;

    /* Setup dial box variables */
    strcpy(dial_box->name, name);
    dial_box->x           = x;
    dial_box->y           = y;
    dial_box->window      = LUI_DialBoxWindow;
    dial_box->border_size = border;

    /* Dimension the dial box label */
    LUI_DialBox->label_x = 0;
    LUI_DialBox->label_y = 0;

    /* Add dial box to head of list */
    dial_box->np = LUI_DialBoxHead;
    LUI_DialBoxHead  = dial_box;

    /* Add Expose event */
    LUI_EventAdd(LUI_DialBoxWindow, ExposureMask, LUI_DialBoxDraw);

    return(1);
}

/******************************************************************************/
LUI_DialBoxClose(name)
    char *name;
{
    int x, y, width, height, bs;

    /* Check for errors */
    if (!LUI_DialBox || strcmp(LUI_DialBox->name, name)) {
	fprintf(stderr, "LUI_DialBoxClose: Can't find open DialBox %s!\n", name);
	return(0);
    }

    /* Set width and height */
    bs = LUI_DialBox->border_size;
    LUI_DialBox->width  = LUI_DialBoxWidth  + bs;
    LUI_DialBox->height = LUI_DialBoxHeight + bs;

    /* Reset size of window and reposition it */
    x      = LUI_DialBox->x;
    y      = LUI_DialBox->y;
    width  = LUI_DialBox->width;
    height = LUI_DialBox->height;
    XResizeWindow(LUI_Display, LUI_DialBoxWindow, width, height);
    XMoveWindow(LUI_Display, LUI_DialBoxWindow, x, y);

#ifdef LUI_DIALBOX_LABEL
    /* Create the label */
    LUI_DialBox->label_text[0] = LUI_DialBox->name;
    LUI_DialBox->label = 
      LUI_LabelOpen(name, 
		1, LUI_DialBox->label_text,
		LUI_DialBoxWindow, 
		bs, bs, width - 2*bs, LUI_DialBox->label_y, 
		4, NULL);
    LUI_LabelClose(name);
    LUI_LabelColor(name, LUI_DefaultHue2);
    LUI_LabelHighlight(name, 0);
#endif

    /* Reset LUI_Dial Globals */
    LUI_DialBox = NULL;

    return(1);
}

/******************************************************************************/
LUI_DialBoxVisible(name, on_off)
    char *name;
    int on_off;
{
    register LUI_DIAL_BOX *dp;

    for (dp = LUI_DialBoxHead; dp; dp = dp->np) 
      if (!strcmp(dp->name, name)) break;

    if (dp) {
	if (on_off) {
	  XMapWindow  (LUI_Display, dp->window);
#ifdef LUI_DIALBOX_LABEL
	  LUI_LabelVisible(dp->name, 1);
#endif
	  dp->visible = 1;
        }
	else {
	  XUnmapWindow(LUI_Display, dp->window);
#ifdef LUI_DIALBOX_LABEL
	  LUI_LabelVisible(dp->name, 0);
#endif
	  dp->visible = 0;
        }
	return(1);
    }
    else {
	fprintf(stderr, "LUI_DialBoxVisible: can't find DialBox %s\n", name);
	return(0);
    }
}

/******************************************************************************/
LUI_DialBoxMove(name, x, y)
    char *name;
    int x, y;
{
    register LUI_DIAL_BOX *dp;
    
    for (dp = LUI_DialBoxHead; dp; dp = dp->np)
      if (!strcmp(dp->name, name)) break;

    if (dp) {
	XMoveWindow(LUI_Display, dp->window, x, y);
	return(1);
    }
    else {
	fprintf(stderr, "LUI_DialBoxMove: can't find DialBox %s\n", name);
	return(0);
    }
}

/******************************************************************************/
LUI_DialBoxQuery(name, x, y, width, height)
    char *name;
    int *x, *y, *width, *height;
{
    register LUI_DIAL_BOX *dp;

    for (dp = LUI_DialBoxHead; dp; dp = dp->np) 
      if (!strcmp(dp->name, name)) break;

    if (dp) {
	*x = dp->x;
	*y = dp->y;
	*width = dp->width;
	*height = dp->height;
	return(1);
    }
    else {
	fprintf(stderr, "LUI_DialBoxQuery: can't find DialBox %s\n", name);
	return(0);
    }
}

/******************************************************************************/
 LUI_DIAL_BOX *
LUI_DialBoxLookupWindow(window)
    Window window;
{
    register LUI_DIAL_BOX *dp;

    for (dp = LUI_DialBoxHead; dp; dp = dp->np) 
	if (dp->window == window) 
	  return(dp);
    return(NULL);
}

/******************************************************************************/
 LUI_DIAL_BOX *
LUI_DialBoxLookupName(name)
    char *name;
{
    register LUI_DIAL_BOX *dp;

    for (dp = LUI_DialBoxHead; dp; dp = dp->np) 
	if (!(strcmp(name, dp->name))) 
	  return(dp);
    return(NULL);
}
   

/******************************************************************************/
LUI_DialBoxDraw(window, event)
    Window window;
    XEvent *event;
{
    LUI_DIAL_BOX *dp;

    /* Squelch multiple redraws */
    if (event->xexpose.count > 0)
      return;

    if ((dp = LUI_DialBoxLookupWindow(event->xexpose.window)) == NULL)
      return;

    LUI_DrawBorderAt(window, 0, 0, dp->width, dp->height, dp->border_size, 1);
}

/******************************************************************************/
LUI_DialCreate(label, type, x, y, low, high, init, func)
    char *label;
    int type, x, y;
    float low, high, init;
    LUI_FNCP func;
{
    register LUI_DIAL *dial;
    Pixmap LUI_DialMakePixmap();
    Window window;
    XSetWindowAttributes window_attributes;
    int xoff, yoff, w, h, LUI_DialProcess();

    /* Adjust position based on LUI_Dial size */
    x = (x * LUI_DIAL_X_SIZE) + LUI_DialBox->border_size;
#ifdef LUI_DIALBOX_LABEL
    y = (y * LUI_DIAL_Y_SIZE) + LUI_DialBox->border_size + LUI_DialBox->label_y;
#endif
    y = (y * LUI_DIAL_Y_SIZE) + LUI_DialBox->border_size - 1;

    /* Adjust max width and height of LUI_Dial Box window */
    xoff = LUI_DIAL_X_SIZE;
    yoff = LUI_DIAL_Y_SIZE;
    if ((x + xoff) > LUI_DialBoxWidth ) LUI_DialBoxWidth  = x + xoff;
    if ((y + yoff) > LUI_DialBoxHeight) LUI_DialBoxHeight = y + yoff;

    /* Create the dial window */
    window_attributes.background_pixel = LUI_Color_FG;
    window = XCreateWindow(LUI_Display, 
			   LUI_DialBoxWindow, 
			   x, y, 
			   LUI_DIAL_X_SIZE, LUI_DIAL_Y_SIZE,
			   0,
			   LUI_Depth,
			   InputOutput,
			   CopyFromParent,
			   CWBackPixel,
			   &window_attributes);

    /* Add into Event Table */
    LUI_EventAdd(window, 
		 ExposureMask | EnterWindowMask | 
		 LeaveWindowMask | ButtonPressMask | ButtonReleaseMask, 
		 LUI_DialProcess);

    LUI_EventAdd(window, ButtonPressMask | ButtonReleaseMask, LUI_DialFunction);

    /* Allocate a new dial */
    if (!(dial = (LUI_DIAL *)malloc(sizeof(LUI_DIAL)))) {
	fprintf(stderr, "LUI_DialCreate: can't allocate new dial\n");
	exit(-1);
    }

    /* Initialize dial */
    strcpy(dial->label, label);
    dial->state      = LUI_DIAL_INACTIVE;
    dial->type       = type;
    dial->x          = x;
    dial->y          = y;
    dial->width      = LUI_DIAL_X_SIZE;
    dial->height     = LUI_DIAL_Y_SIZE;
    dial->func       = func;
    dial->window     = window;
    dial->dial_box   = LUI_DialBox;
    dial->low        = low;
    dial->high       = high;
    dial->value      = init;
    dial->init       = init;
    dial->last_angle = 0.0;

    /* Define dial pixmaps */
    dial->pixmap_in  = LUI_ReadPixmap(LUI_DIAL_PIXMAP_IN,  0, 0, &w, &h);
    dial->pixmap_out = LUI_ReadPixmap(LUI_DIAL_PIXMAP_OUT, 0, 0, &w, &h);

    LUI_DrawBorderAt(dial->pixmap_in, 0, 0, dial->width, dial->height, 2, 1);
    DrawTicks(dial, dial->pixmap_in,  8, 15);
    DrawTicks(dial, dial->pixmap_in, 16, 10);
    DrawTicks(dial, dial->pixmap_in, 32,  5);

    /* Add dial to head of list */
    dial->np = LUI_DialHead;
    LUI_DialHead = dial;

    /* Map the dial window */
    XMapWindow(LUI_Display, window);

    return(1);
}

/******************************************************************************/
LUI_DialDraw(dial, value, delta)
    register LUI_DIAL *dial;
    float value, delta;
{
    int x0, y0, x1, y1, x2, y2, x3, y3, cx, cy, size;
    char label[64];
    float angle;
    XPoint arrow[4];
    static float valufac = 4.0 / 3.0;

    angle = value * 360.0;
    angle -= 90.0;
    if (angle < 0)
      angle += 360;

    if (dial->type == LUI_DIAL_ACCUMULATOR) {
	if (delta != 0.0)
	  dial->value += delta * dial->high;
    }
    else {
	if (delta != 0.0)
	  dial->value += delta * valufac * (dial->high - dial->low);
    }
 
    cx = dial->width / 2;
    cy = dial->height - dial->width / 2;

    XCopyArea(LUI_Display, dial->pixmap_in, dial->pixmap_out, LUI_GC_fore, 
	      0, 0, dial->width, dial->height, 0, 0);

    x0 = (40 * cos(DERAD(angle   ))) + cx + 0.5;
    y0 = (40 * sin(DERAD(angle   ))) + cy + 0.5;
    
    x1 = (10 * cos(DERAD(angle-90))) + cx + 0.5;
    y1 = (10 * sin(DERAD(angle-90))) + cy + 0.5;
    
    x2 = (10 * cos(DERAD(angle   ))) + cx + 0.5;
    y2 = (10 * sin(DERAD(angle   ))) + cy + 0.5;
    
    x3 = (10 * cos(DERAD(angle+90))) + cx + 0.5;
    y3 = (10 * sin(DERAD(angle+90))) + cy + 0.5;
    
    arrow[0].x = x0; arrow[0].y = y0;
    arrow[1].x = x1; arrow[1].y = y1;
    arrow[2].x = x2; arrow[2].y = y2;
    XFillPolygon(LUI_Display, dial->pixmap_out, LUI_GC_dk_grey, arrow, 3, Nonconvex, 
		 CoordModeOrigin);
    
    arrow[0].x = x0; arrow[0].y = y0;
    arrow[1].x = x3; arrow[1].y = y3;
    arrow[2].x = x2; arrow[2].y = y2;
    XFillPolygon(LUI_Display, dial->pixmap_out, LUI_GC_lt_grey, arrow, 3, Nonconvex, 
		 CoordModeOrigin);
    
    arrow[0].x = x3; arrow[0].y = y3;
    arrow[1].x = x1; arrow[1].y = y1;
    arrow[2].x = x2; arrow[2].y = y2;
    XFillPolygon(LUI_Display, dial->pixmap_out, LUI_GC_sl_grey, arrow, 3, Nonconvex, 
		 CoordModeOrigin);

    /* Draw the dial label */
    cy = 12;
	
    if (dial->state == LUI_DIAL_ACTIVE) 
      sprintf(label, "%5.2f", dial->value);	
    else 
      sprintf(label, "%s", dial->label);
    size  = strlen(label);
	
    cx -= XTextWidth(LUI_Font, label, size) / 2.0 + 0.5;
    cy += LUI_Font_yoff / 2.0 + 0.5;
	
    XDrawString(LUI_Display, dial->pixmap_out, LUI_GC_fore, cx, cy, label, size);

    /* Exec the dial function */
    if (dial->func) 
      (*(dial->func))(dial, delta);

    /* Change the window's background pixmap according to dial state */
    XSetWindowBackgroundPixmap(LUI_Display, dial->window, dial->pixmap_out);

    /* Clear the dial window (forces a redraw) */
    XClearWindow(LUI_Display, dial->window);
}

/******************************************************************************/
DrawTicks(dial, pixmap, n, thick)
    LUI_DIAL *dial;
    Pixmap pixmap;
    int n, thick;
{
    register int i;
    int x0, y0, x1, y1, cx, cy;
    float angle, anginc;

    cx = dial->width / 2;
    cy = dial->height - dial->width / 2;

    anginc = 360.0 / n;

    for (i = 0; i < n; ++i) {
	angle = anginc * i;

	if (dial->type == LUI_DIAL_VALUATOR && angle > 45 && angle < 135) 
	  continue;

	x0 = (40 * cos(DERAD(angle))) + cx + 0.5;
	y0 = (40 * sin(DERAD(angle))) + cy + 0.5;
	
	x1 = ((40 - thick) * cos(DERAD(angle))) + cx + 0.5;
	y1 = ((40 - thick) * sin(DERAD(angle))) + cy + 0.5;
	
	XDrawLine(LUI_Display, pixmap, LUI_GC_fore,
		  x0, y0, x1, y1);
    }
}

/******************************************************************************/
 LUI_DIAL *
LUI_DialLookupWindow(window)
    Window window;
{
    register LUI_DIAL *dp;

    for (dp = LUI_DialHead; dp; dp = dp->np) 
	if (dp->window == window) return(dp);

    return(NULL);
}

/******************************************************************************/
LUI_DialHighlight(dial, on)
    register LUI_DIAL *dial;
    register int on;
{
    if (on) {
	LUI_Cursor(dial->window, LUI_DIAL_CURSOR);
#ifdef HIGHLIGHT
	XDrawRectangle(LUI_Display, dial->window, LUI_GC_xor,
		       1, 1, LUI_DIAL_X_SIZE - 2, LUI_DIAL_Y_SIZE - 2);
	XDrawRectangle(LUI_Display, dial->window, LUI_GC_xor,
		       2, 2, LUI_DIAL_X_SIZE - 4, LUI_DIAL_Y_SIZE - 4);
#endif
    }
    else {
	LUI_Cursor(dial->window, LUI_MAIN_CURSOR);
#ifdef HIGHLIGHT
	LUI_DialDraw(dial);
#endif
    }
}

/******************************************************************************/
 static 
LUI_DialProcess(window, event)
    Window  window;
    XEvent *event;
{
    Window dummy_window;
    LUI_LABEL *label;
    LUI_DIAL *dial;
    int cursor, x,y, dummy_x,dummy_y, cx,cy, pinned;
    unsigned int mask;
    static int done, oy,ox, last_delta_x, last_delta_y;
    float delta_x, delta_y, delta_angle, last_angle;
    static float angle;

    /* Get the dial function */
    if (!(dial = LUI_DialLookupWindow(window))) {
	fprintf(stderr, "LUI_DialProcess: can't find dial for window %d\n", window);
	return;
    }
#ifdef LUI_DIALBOX_LABEL
    label = dial->dial_box->label;
#endif

    /* Process the event */
    switch (event->type) {
#ifdef LUI_DIALBOX_LABEL
      case EnterNotify:
	LUI_DialHighlight(dial, 1);
	LUI_LabelHighlight(label->name, 1);
	break;
	
      case LeaveNotify:
	LUI_DialHighlight(dial, 0);
	LUI_LabelHighlight(label->name, 0);
	break;
#endif

      case Expose:
	if (event->xexpose.count == 0) 
	  LUI_DialDraw(dial, dial->value, 0);
	break;

      case ButtonPress:
	dial->state = LUI_DIAL_ACTIVE;

#ifdef OLD
	if (event->xbutton.button == Button1) 
	  cursor = LUI_DIAL_CCW_CURSOR;
	else if (event->xbutton.button == Button2)
	  cursor = LUI_DIAL_CW_CURSOR;
	else {
	    dial->value = dial->init;
	    cursor = LUI_DIAL_RES_CURSOR;
	    LUI_Cursor(window, cursor);
	    dial->last_angle = 0.0;
	    LUI_DialDraw(dial, 0.0, 0.0);

	    if (dial->func) 
	      (*(dial->func))(dial);
	    break;
	}
#endif
	if (event->xbutton.button == Button1) 
	  cursor = LUI_DIAL_CCW_CURSOR;
	else if (event->xbutton.button == Button2)
	  cursor = LUI_DIAL_RES_CURSOR;
	else 
	  cursor = LUI_DIAL_CW_CURSOR;

	/* Set the cursor */
	LUI_Cursor(window, cursor);

	cx = dial->width  / 2;
	cy = dial->height / 2;

	XTranslateCoordinates(LUI_Display, window, LUI_RootWindow, cx, cy,
			      &ox, &oy, &dummy_window);

	last_delta_x = last_delta_y = -999;
	last_angle = dial->last_angle;

	pinned = 0;

	for (done = 0; !done; ) {
	    if (XPending(LUI_Display)) {
		XEvent event;

		XPeekEvent(LUI_Display, &event);
		if (event.type == ButtonRelease) {
		    XNextEvent(LUI_Display, &event);

		    LUI_Cursor(window, LUI_DIAL_CURSOR);
		    dial->state = LUI_DIAL_INACTIVE;
		    if (dial->value == dial->init)
		      LUI_DialDraw(dial, 0.0, 0.0);
		    else
		      LUI_DialDraw(dial, angle/360.0, 0.0);
		    break;
		}
		else {
		    XNextEvent(LUI_Display, &event);
/*
		    LUI_EventDispatch(&event);
*/
		}
	    }
	    XQueryPointer(LUI_Display, window, &dummy_window, &dummy_window, &x, &y,
			  &dummy_x, &dummy_y, &mask);

	    if (event->xbutton.button == Button2) {
		/* Compute the delta offsets */
		delta_x = x - ox;
		delta_y = y - oy;
		
		/* Continue if too close to center */
		if ((fabs(delta_x) < 15) && (fabs(delta_y) < 15))
		  continue;
		
		/* Continue if haven't moved */
		if ((delta_x == last_delta_x) && (delta_y == last_delta_y))
		  continue;
		
		/* Compute the angle.  This is the absolute angle on */
		/* the screen, based on 12 o'clock being 0 degrees   */
		angle = RADDE(atan2(delta_y, delta_x));
		angle += 90;
		if (angle < 0.0) 
		  angle += 360.0;
		else if (angle > 360.0) 
		  angle -= 360.0;
		
		if (dial->type == LUI_DIAL_VALUATOR) {
		    if (pinned == 1) {
			if (angle < 135) 
			  pinned = 0;
			else {
			    angle = 135;
			    continue;
			}
		    }
		    else if (pinned == 2) {
			if (angle > 225) 
			  pinned = 0;
			else {
			    angle = 225;
			    continue;
			}
		    }
		    if (angle > 135 && angle < 225) {
			if (angle > 135 && angle < 180) {
			    pinned = 1;
			    angle = 135;
			}
			else if (angle < 225 && angle > 180) {
			    pinned = 2;
			    angle = 225;
			}
		    }
		}
	    }
	    else if (event->xbutton.button == Button1) 
		angle = last_angle - (dial->high - dial->low) * .05;
	    else if (event->xbutton.button == Button3) 
		angle = last_angle + (dial->high - dial->low) * .05;
		
	    /* Compute the delta angle.  This will be the amount */
	    /* the dial moved relative to the last position      */
	    delta_angle = angle - last_angle;
	    if (delta_angle > 180.0) 
	      delta_angle -= 360.0;
	    else if (delta_angle < -180.0) 
	      delta_angle += 360.0;
#ifdef DEBUG
	    fprintf(stderr, "angle = %c%05.2f  delta_angle = %c%04.2f\r", 
		    angle > 0 ? ' ' : '-', fabs(angle), 
		    delta_angle > 0 ? ' ' : '-', fabs(delta_angle));
#endif
	    /* Redraw the dial */
	    LUI_DialDraw(dial, angle/360.0, delta_angle/360.0);

	    /* Reset last values */
	    last_delta_x = delta_x;
	    last_delta_y = delta_y;
	    last_angle   = angle;
	    dial->last_angle = last_angle;

	    if (dial->func) 
	      (*(dial->func))(dial);
	}
	break;

      case ButtonRelease:
	break;
    }
}

